// ****************************************************************************
// Projekt      : LCD Module DOGM163S-A
// Prozessor    : PIC18F45K22 @ 40Mhz
// Compiler     : MPLAB XC8 V2.45
// Autor        : Stefan Gluekler
// Datum        : 2023.12.18
// Version      : V2.1
// Beschreibung : Ansteurung des LCD-Moduls DOGM163S-A
//                und dem "VFEI - PIC Board V2015"
// ****************************************************************************

// -- I N C L U D E S ---------------------------------------------------------

#include "LCD_Module_DOGM163S-A.h"

#include <xc.h>             // Include for the XC8 compiler
#include <stdbool.h>        // Include for data type 'bool' and keywords 'true' and 'false'
#include <stdint.h>         // Include for defines the standard integer types per ISO/IEC 9899:1999

#include <string.h>         // Include for functions strcpy,strlen,...
#include <stdio.h>          // Include for functions sprintf,...

// -- D E F I N E S -----------------------------------------------------------

#define _XTAL_FREQ 40000000     // Crystal oscillator frequencies of 10Mhz VFEI-PIC-Board 2012 and 4X PLL
                                // Required by __delay_xx()

#define SPI_CS_TRIS       TRISCbits.TRISC4
#define SPI_CS            LATCbits.LATC4
#define SPI_CS_OFF        SPI_CS = 1
#define SPI_CS_ON         SPI_CS = 0

// SPI
#define SPI_FOSC_64       0b00000010            // SPI Master mode, clock = Fosc/64
#define MODE_00           0b00000000            // Setting for SPI bus Mode 0,0
#define SMPEND            0b10000000            // Input data sample at end of data out             
#define SSPENB            0b00100000            // Enable serial port and configures SCK, SDO, SDI

// Timer2
#define TIMER_INT_OFF     0b00000000            // Disable TIMER Interrupt
#define T2_PS_1_1         0b00000000            // Timer2 Prescale 1:1
#define T2_POST_1_8       0b00111000            // Timer2 Postscaler 1:8

// PWM1
#define ECCP_1_SEL_TMR12  0b00000000            // ECCP selects TIMER1 for Capture & Compare and TIMER2 for PWM
#define ECCP1CON          CCP1CON

#define LCD_MODE          LATBbits.LATB5
#define LCD_MODE_DATA     LCD_MODE = 1
#define LCD_MODE_COMMAND  LCD_MODE = 0

#define LCD_COMMAND_DELAY_SHORT __delay_us(500)     // Short commands
#define LCD_COMMAND_DELAY_LONG  __delay_us(1000)    // Long commands
#define LCD_DATA_DELAY          __delay_us(2)       // Data

#pragma warning disable 520 // Ausschalten Warnungen von nicht benutzten Funktionen

// -- G L O B A L E  V A R I A B L E N ----------------------------------------

// -- F U N K T I O N E N - P R O T O T Y P E S -------------------------------

void LCD_Module_OpenTimer2(uint8_t config);

void LCD_Module_OpenEPWM1(uint8_t period,
                          uint8_t timer_source );
void LCD_Module_SetDCEPWM1(uint16_t dutycycle);

void LCD_Module_OpenSPI1(uint8_t sync_mode,
                         uint8_t bus_mode,
                         uint8_t smp_phase);
signed char LCD_Module_WriteSPI(uint8_t data_out);

// -- F U N K T I O N E N -----------------------------------------------------

// Initialisieren des LCD-Moduls
void LCD_Module_Init(void)
{
  TRISBbits.TRISB5 = 0; // PortB[5]: Output port
  ANSELBbits.ANSB5 = 0; // PortB[5]: Digital-IO
  LATBbits.LATB5   = 0; // PortB[5]: Mode command

  TRISCbits.TRISC2 = 0; // PortC[2]: Output port
  ANSELCbits.ANSC2 = 0; // PortC[2]: Digital-IO
  LATCbits.LATC2   = 0; // PortC[2]: LCD Hintergrund-Beleuchtung aus

  TRISCbits.TRISC3 = 0; // PortC[3]: Output port
  ANSELCbits.ANSC3 = 0; // PortC[3]: Digital-IO
  LATCbits.LATC3   = 0; // PortC[3]: LCD Clock aus

  TRISCbits.TRISC4 = 0; // PortC[4]: Output port
  ANSELCbits.ANSC4 = 0; // PortC[4]: Digital-IO
  LATCbits.LATC4   = 0; // PortC[4]: LCD SDO aus

  TRISCbits.TRISC5 = 0; // PortC[5]: Output port
  ANSELCbits.ANSC5 = 0; // PortC[5]: Digital-IO
  LATCbits.LATC5   = 0; // PortC[5]: LCD SDI aus

  // Init Timer 2 (for PWM 1)
  LCD_Module_OpenTimer2( TIMER_INT_OFF &
                         T2_PS_1_1     &
                         T2_POST_1_8     );

  // Init PWM 1 (on Timer 2)
  LCD_Module_OpenEPWM1( 0xFF ,
                        ECCP_1_SEL_TMR12 );

  // Set backlight to 0%
  LCD_Module_SetBacklight( 0 );

  // Init SPI Fosc/64, in mode 0, master mode
  LCD_Module_OpenSPI1( SPI_FOSC_64,
                       MODE_00,
                       SMPEND );

  // Konfiguration der Ports
  LCD_MODE_COMMAND;             // Mode command
  SPI_CS_TRIS = 0;              // Output port
  SPI_CS_OFF;                   // Chip select aus
  LCD_COMMAND_DELAY_LONG;       // Lange Zeit warten

  // Konfiguration des Display
  LCD_MODE_COMMAND;             // Mode command
  SPI_CS_ON;                    // Chip select on
  LCD_Module_WriteSPI( 0x39 );  // Function set; 8 bit Datenlnge, 2 Zeilen, Instruction table 1
  LCD_COMMAND_DELAY_SHORT;      // Kurze Zeit warten
  LCD_Module_WriteSPI( 0x1D );  // Bias Set; BS 1/4; 3 zeiliges Display
  LCD_COMMAND_DELAY_SHORT;      // Kurze Zeit warten
  LCD_Module_WriteSPI( 0x50 );  // Booster aus; Kontrast C5, C4 setzen
  LCD_COMMAND_DELAY_SHORT;      // Kurze Zeit warten
  LCD_Module_WriteSPI( 0x70 );  // C0 bic C3 setzen
  LCD_COMMAND_DELAY_SHORT;      // Kurze Zeit warten
  LCD_Module_WriteSPI( 0x6C );  // Spannungsfolger und Verstrkung setzen
  LCD_COMMAND_DELAY_SHORT;      // Kurze Zeit warten
  LCD_Module_WriteSPI( 0x78 );  // Kontrast C3 setzen
  LCD_COMMAND_DELAY_SHORT;      // Kurze Zeit warten
  LCD_Module_WriteSPI( 0x38 );  // Function set; 8 bit Datenlnge, 2 Zeilen, Instruction table 0
  LCD_COMMAND_DELAY_SHORT;      // Kurze Zeit warten
  LCD_Module_WriteSPI( 0x08 );  // Display EIN
  LCD_COMMAND_DELAY_SHORT;      // Kurze Zeit warten
  LCD_Module_WriteSPI( 0x01 );  // Display lschen, Cursor Home
  LCD_COMMAND_DELAY_LONG;       // Lange Zeit warten
  LCD_Module_WriteSPI( 0x06 );  // Cursor Auto-Increment
  LCD_COMMAND_DELAY_SHORT;      // Kurze Zeit warten
  SPI_CS_OFF;                   // Chip select off

  // Set Cursor off
  LCD_Module_SetCursorMode(false,
                           false);

  // Clear LCD
  LCD_Module_ClearAll();

  // Set Backlight to 100%
  LCD_Module_SetBacklight(100);
  // Set Contrast to 15%
  LCD_Module_SetContrast(15);
}

// Setzen der Helligkeit der Hintergrund-Beleuchtung
// percent  -> Helligkeit der Hintergrund-Beleuchtung (0-100)
void LCD_Module_SetBacklight(uint8_t percent)
{
  uint16_t pwm_value;
  if (percent > 100)
      percent = 100;
  pwm_value = (uint16_t)((0x3FF * (uint32_t)percent) / 100);
  LCD_Module_SetDCEPWM1( pwm_value );
}

// Setzen des LCD Kontrastes
// percent  -> Kontrast des LCD (0-100)
void LCD_Module_SetContrast(uint8_t percent)
{
  uint8_t contrast,contrast_high,contrast_low;
  if (percent > 100)
      percent = 100;
  // Berechnen des Kontrast-Wertes fr LCD (C0-C5 - 6Bit)
  contrast = (uint8_t)((0x3F * (uint32_t)percent) / 100);
  contrast_high = (contrast & 0b00110000) >> 4; // C5 und C4 ausmaskieren und nach recht schieben
  contrast_low  = (contrast & 0b00001111);      // C0 bis C3 ausmaskieren
  // Konfiguration des Display
  LCD_MODE_COMMAND;             // Mode command
  SPI_CS_ON;                    // Chip select on
  LCD_Module_WriteSPI( 0x39 );  // Function set; 8 bit Datenlnge, 2 Zeilen, Instruction table 1
  LCD_COMMAND_DELAY_SHORT;      // Kurze Zeit warten
  LCD_Module_WriteSPI( 0x50 | contrast_high );  // Booster aus; C5 und C4 setzen
  LCD_COMMAND_DELAY_SHORT;      // Kurze Zeit warten
  LCD_Module_WriteSPI( 0x70 | contrast_low );   // C0 bic C3 setzen
  LCD_COMMAND_DELAY_SHORT;      // Kurze Zeit warten
  LCD_Module_WriteSPI( 0x38 );  // Function set; 8 bit Datenlnge, 2 Zeilen, Instruction table 0
  LCD_COMMAND_DELAY_SHORT;      // Kurze Zeit warten
  SPI_CS_OFF;                   // Chip select off
}

// Loeschen aller Zeilen
void LCD_Module_ClearAll(void)
{
  LCD_MODE_COMMAND;             // Mode command
  SPI_CS_ON;                    // Chip select on
  LCD_Module_WriteSPI( 0x01 );  // Display loeschen, Cursor Home
  LCD_COMMAND_DELAY_LONG;       // Lange Zeit warten
  SPI_CS_OFF;                   // Chip select off
}

// Loeschen einer Zeile
// line     -> Zeile des Curors (1-3)
void LCD_Module_ClearLine(uint8_t line)
{
  LCD_Module_SetCursorPosition(1,line);
  LCD_Module_WriteString("                ");
  LCD_Module_SetCursorPosition(1,line);
}

// Setzen des Cursor-Mode
// cursorOn    = TRUE   -> Cursor ist sichtbar
// cursorOn    = FALSE  -> Cursor ist unsichtbar
// cursorBlink = TRUE   -> Cursor blinkt
// cursorBlink = FALSE  -> Cursor ist fix
void LCD_Module_SetCursorMode(bool cursorOn,
                              bool cursorBlink)
{
  uint8_t command = 0b00001100; // Display on

  if (cursorOn == true)
    command |= 0b00000010;
  if (cursorBlink == true)
    command |= 0b00000001;

  LCD_MODE_COMMAND;               // Mode command
  SPI_CS_ON;                      // Chip select on
  LCD_Module_WriteSPI( command ); // Display EIN
  LCD_COMMAND_DELAY_SHORT;        // Kurze Zeit warten
  SPI_CS_OFF;                     // Chip select off
}

// Setzen der Cursor-Position auf Position und Zeile
// position -> Position des Cursors (1-16)
// line     -> Zeile des Curors (1-3)
void LCD_Module_SetCursorPosition(uint8_t position,
                                  uint8_t line)
{
  uint8_t command = 0b10000000; // Set DRAM Position

  // Ungueltige Position
  if (position < 1 || position > LCD_CHARACHTERS)
    return;
  // Ungueltige Linie
  if (line < 1 || line > LCD_LINES)
    return;
    
  command |= (0x10 * (line-1)) + (position-1);

  LCD_MODE_COMMAND;               // Mode command
  SPI_CS_ON;                      // Chip select on
  LCD_Module_WriteSPI( command ); // Display EIN
  LCD_COMMAND_DELAY_SHORT;        // Kurze Zeit warten
  SPI_CS_OFF;                     // Chip select off
}

// Ausgabe eines Zeichen
// charachter  ->  Zeichen dies ausgegeben werden soll (char)
void LCD_Module_WriteChar(char charachter)
{
  LCD_MODE_DATA;                  // Mode data
  SPI_CS_ON;                      // Chip select on
  LCD_Module_WriteSPI( charachter );
  LCD_DATA_DELAY;                 // Kurze Zeit warten
  SPI_CS_OFF;                     // Chip select off
}

// Ausgabe eines Strings
// romString   ->  String der ausgegeben werden soll (char[])
void LCD_Module_WriteString(const char *romString)
{
  uint16_t stringPosition;
  uint16_t stringLen;

  stringLen = (uint16_t)strlen(romString);

  LCD_MODE_DATA;                // Mode data
  SPI_CS_ON;                    // Chip select on
  for (stringPosition=0;stringPosition<stringLen;stringPosition++)
  {
    LCD_Module_WriteSPI( romString[stringPosition] );
    LCD_DATA_DELAY;             // Kurze Zeit warten
  } // for()
  SPI_CS_OFF;                   // Chip select off
}

// Formatierte Ausgabe eines Strings
// romString   ->  String der ausgegeben werden soll (char[])
// romFormat  ->  Gewuenschtes Ausgabeformat
void LCD_Module_WriteFormated_String(const char *romString,
                                     const char *romFormat)
{
  char buffer[LCD_CHARACHTERS+1];
  
  sprintf(buffer,romFormat,romString);
  LCD_Module_WriteString(buffer);  
}

// Formatierte Ausgabe einer int8-/int16-Variable
// value      ->  Wert der ausgegeben werden soll
// romFormat  ->  Gewuenschtes Ausgabeformat
void LCD_Module_WriteFormated_Int16(int16_t value,
                                    const char *romFormat)
{
  char buffer[LCD_CHARACHTERS+1];
  
  sprintf(buffer,romFormat,value);
  LCD_Module_WriteString(buffer);
}

// Formatierte Ausgabe einer uint8/uint16-Variable
// value      ->  Wert der ausgegeben werden soll
// romFormat  ->  Gewuenschtes Ausgabeformat
void LCD_Module_WriteFormated_UInt16(uint16_t value,
                                     const char *romFormat)
{
  char buffer[LCD_CHARACHTERS+1];

  sprintf(buffer,romFormat,value);
  LCD_Module_WriteString(buffer);
}

// Formatierte Ausgabe einer int32-Variable
// value      ->  Wert der ausgegeben werden soll
// romFormat  ->  Gewuenschtes Ausgabeformat
void LCD_Module_WriteFormated_Int32(int32_t value,
                                    const char *romFormat)
{
  char buffer[LCD_CHARACHTERS+1];
  
  sprintf(buffer,romFormat,value);
  LCD_Module_WriteString(buffer);
}

// Formatierte Ausgabe einer uint32-Variable
// value      ->  Wert der ausgegeben werden soll
// romFormat  ->  Gewuenschtes Ausgabeformat
void LCD_Module_WriteFormated_UInt32(uint32_t value,
                                     const char *romFormat)
{
  char buffer[LCD_CHARACHTERS+1];

  sprintf(buffer,romFormat,value);
  LCD_Module_WriteString(buffer);
}

// Timer2: Initialisation
void LCD_Module_OpenTimer2(uint8_t config)
{
  T2CON = (0xfb & config);  // Set all configuration values, but
                            // don't start timer yet 								
  TMR2 = 0;                 // Clear Timer2
  PIR1bits.TMR2IF = 0;
  if(config & 0x80)         // Enable timer interrupts?
    PIE1bits.TMR2IE = 1;
  else
    PIE1bits.TMR2IE = 0;
  T2CONbits.TMR2ON = 1; // Turn on Timer2
}

// PWM1: Initialisation
void LCD_Module_OpenEPWM1(uint8_t period,
                          uint8_t timer_source )
{
  ECCP1CON=0b00001100;   // ccpxm3:ccpxm0 11xx=pwm mode
  // configure timer source for CCP
  CCPTMRS0 &= 0b11111100;
  CCPTMRS0 |= ((timer_source&0b00110000)>>4);    
  if( (CCPTMRS0&0x03)==0x00)
  {
    T2CONbits.TMR2ON = 0;  // STOP TIMERx registers to POR state
    PR2 = period;          // Set period
    T2CONbits.TMR2ON = 1;  // Turn on PWMx
  }
  else if( (CCPTMRS0&0x03)==0x01)
  {
    T4CONbits.TMR4ON = 0;  // STOP TIMERx registers to POR state
    PR4 = period;          // Set period
    T4CONbits.TMR4ON = 1;  // Turn on PWMx
  }
  else if( (CCPTMRS0&0x03)==0x02)
  {
    T6CONbits.TMR6ON = 0;  // STOP TIMERx registers to POR state
    PR6 = period;          // Set period
    T6CONbits.TMR6ON = 1;  // Turn on PWMx  
  }
}

// PWM1: Setzen Wert (0-127))
void LCD_Module_SetDCEPWM1(uint16_t dutycycle)
{
  union PWMDC
  {
    uint16_t lpwm;
    uint8_t bpwm[2];
  };
  union PWMDC DCycle;
  // Save the dutycycle value in the union
  DCycle.lpwm = dutycycle << 6;
  // Write the high byte into ECCPR1L
  CCPR1L = DCycle.bpwm[1];
  // Write the low byte into ECCP1CON5:4
  ECCP1CON = (ECCP1CON & 0xCF) | ((DCycle.bpwm[0] >> 2) & 0x30);
}

// SPI: Initialisation
void LCD_Module_OpenSPI1(uint8_t sync_mode,
                         uint8_t bus_mode,
                         uint8_t smp_phase)
{
  SSP1STAT &= 0x3F;               // power on state 
  SSP1CON1 = 0x00;                // power on state
  SSP1CON1 |= sync_mode;          // select serial mode 
  SSP1STAT |= smp_phase;          // select data input sample phase
  switch( bus_mode )
  {
    case 0:                       // SPI1 bus mode 0,0
      SSP1STATbits.CKE = 1;       // data transmitted on rising edge
    break;    
    case 2:                       // SPI1 bus mode 1,0
      SSP1STATbits.CKE = 1;       // data transmitted on falling edge
      SSP1CON1bits.CKP = 1;       // clock idle state high
    break;
    case 3:                       // SPI1 bus mode 1,1
      SSP1CON1bits.CKP = 1;       // clock idle state high
    break;
    default:                      // default SPI1 bus mode 0,1
    break;
  }
  switch( sync_mode )
  {
    case 4:                       // slave mode w /SS enable
      TRISAbits.TRISA5 = 1;       // define /SS1 pin as input
      TRISCbits.TRISC3 = 1;       // define clock pin as input
    break;
    case 5:                       // slave mode w/o /SS enable
      TRISCbits.TRISC3 = 1;       // define clock pin as input
    break;
    default:                      // master mode, define clock pin as output
      TRISCbits.TRISC3 = 0;       // define clock pin as output
    break;
  }
  TRISCbits.TRISC4 = 1;           // define SDI pin as input
  TRISCbits.TRISC5 = 0;           // define SDO pin as output
  SSP1CON1 |= SSPENB;             // enable synchronous serial port
}

// SPI: Schreiben eines Zeichen
signed char LCD_Module_WriteSPI(uint8_t data_out)
{
  uint8_t TempVar;  
  TempVar = SSPBUF;           // Clears BF
  PIR1bits.SSPIF = 0;         // Clear interrupt flag
  SSPCON1bits.WCOL = 0;			//Clear any previous write collision
  SSPBUF = data_out;           // write byte to SSPBUF register
  if ( SSPCON1 & 0x80 )        // test if write collision occurred
   return ( -1 );              // if WCOL bit is set return negative #
  else
   // while( !SSPSTATbits.BF );  // wait until bus cycle complete 
    while( !PIR1bits.SSPIF );  // wait until bus cycle complete  
   return ( 0 );                // if WCOL bit is not set return non-negative#
}
